-- commonly used functions
local remove = table.remove
local min = math.min
local max = math.max
local floor = math.floor
local format = string.format
local gmatch = string.gmatch
local find = string.find
local concat = table.concat
local sub = string.sub

-- font
local font = nil
local fontn = "fonts/FreeMonoBold.ttf"
local fonts = 12
-- messages
local prints = {}
-- traces
local traces = {}
local tracesk = {}
-- active flag
local show = true
-- redraw flag
local redraw = false

local log = {}

--- Enable drawing
function log.show()
  show = true
  if redraw then
    log.redraw()
  end
end

--- Disable drawing
function log.hide()
  show = false
end

--- Clear all messages and traces
function log.clear()
  for i = #prints, 1, -1 do
    prints[i] = nil
  end
  for i = #traces, 1, -1 do
    local k = traces[i]
    tracesk[k] = nil
    traces[i] = nil
  end
end

--- Print one or more arguments, comma separated
-- @param sz First argument
-- @param ... Following arguments (optional)
local args = {}
function log.print(sz, ...)
  -- handle first argument
  sz = tostring(sz) or "nil"
  -- handle following arguments
  local nargs = select('#', ...)
  if nargs > 0 then
    args[1] = sz
    for i = 1, nargs do
      local a = select(i, ...)
      args[i + 1] = tostring(a)
    end
    sz = concat(args, ", ")
    for i = nargs + 1, 1, -1 do
      args[i] = nil
    end
  end
  -- handle new lines
  if find(sz, "[\n]", 1, false) then
    if sub(sz, -1) ~= "\n" then
      sz = sz .. "\n"
    end
    for sz2 in gmatch(sz, "(.-)\n") do
      prints[#prints + 1] = sz2
    end
  else
    prints[#prints + 1] = sz
  end
  -- keep the log size to a reasonable size
  while #prints > 128 do
    remove(prints, 1)
  end
  if redraw then
    log.redraw()
  end
end

--- Print a formatted string
-- @param sz Format string
-- @param ... Format arguments
function log.printf(sz, ...)
  -- handle non-string values
  sz = tostring(sz) or "nil"
  -- handle format parameters
  if select('#', ...) > 0 then
    sz = format(sz, ...)
  end
  log.print(sz)
end

--- Trace a variable
-- @param k Key
-- @param v Value
function log.trace(k, v)
  k = tostring(k) or k
  if v ~= nil then
    -- add trace
    if tracesk[k] == nil then
      traces[#traces + 1] = k
    end
  else
    -- remove trace
    for i = 1, #traces do
      if traces[i] == k then
        remove(traces, i)
        break
      end
    end
  end
  tracesk[k] = tostring(v) or v
  if redraw then
    log.redraw()
  end
end

--- Build all prints and traces in a string (internal)
-- @param rows Number of rows
-- @return String
local out = {}
function log.build(rows)
  -- clean up output buffer
  for i = #out, rows + 1, -1 do
    out[i] = nil
  end
  local i = 1
  -- traces
  -- shown on top, in order of "tracing"
  if #traces > 0 then
    local nrows = min(#traces, rows)
    for j = 1, nrows do
      local k = traces[j]
      local sz = format("%s:%s", k, tracesk[k])
      out[i] = sz
      i = i + 1
    end
  end
  -- log messages
  -- shown below, in order of "printing"
  if #prints > 0 then
    local nrows = min(#prints, rows - #traces)
    local r = max(#prints - nrows, 0)
    for j = 1, nrows do
      out[i] = prints[j + r]
      i = i + 1
    end
  end
  -- todo: break long lines
  return concat(out, "\n")
end

--- Redraw log
function log.redraw()
  if not show then
    return
  end
  -- make sure the display is initialized
  -- make sure the font is loaded
  if font == nil then
    font = love.graphics.newFont(fontn, fonts)
    if font == nil then
      return
    end
  end
  -- transform according to the default viewport
  -- todo: could work with 'love.graphics.origin'
  love.graphics.reset()
  -- figure out extents
  local width, height = love.graphics.getDimensions()
  local fheight = font:getHeight()
  local rows = floor(height/fheight)
  local rheight = height/rows
  -- redraw
  -- add a dark background
  love.graphics.setColor(0, 0, 0, 64)
  love.graphics.rectangle('fill', 0, 0, width, height)
  -- print everything in one call
  local out = log.build(rows)
  -- todo: handle non-loaded UTF-8 characters
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.setFont(font)
  love.graphics.print(out, 0, 0)
end

return log